/* PGPCrack v0.99b
 * by Mark Miller <markm@voicenet.com>
 * A program that can crack PGP conventionally encrypted files or secret keys.
 * Copyright (c) Mark Miller 1996
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include "md5.h"
#include "idea.h"
#include "pgpcrack.h"

void inline chop(char *buf);
void inline hash(char *buf, char *md5hash);
int ctb_check(char c);
int inline ideacrack(char *buf1, char *buf2, char *key);
word16 inline bit2byte(word16 word);
word16 inline byteorder(word16 word);
unsigned char check_alg(FILE *fp);
word16 inline secret_crack(byte *iv, byte p[512], byte q[512], byte d[512],
	byte u[512], word16 p_len, word16 q_len, word16 d_len, word16 u_len, char *key);
int skip_mpi(FILE *fp);
int read_mpi(byte *dat, FILE *fp);
word16 conv_word16(byte *buf);

main(int argc, char *argv[])
{
	FILE *phrases;
	FILE *ideafile;
	FILE *log;
	char md5hash[16];
	char data[256];
	char buf1[10];
	char buf2[8];
	byte p[512], q[512], d[512], u[512];
	char ctb;
	byte alg;
	byte iv[8];
	word16 checksum2 = 0;
	word16 p_len, q_len, d_len, u_len;
	int ctb_return;
	int secret_key = 0;
	if(argc < 3) {
		fprintf(stderr, "Usage: %s [phraselist] [pgpfile] <logfile>\n", argv[0]);
		exit(1);
	}
	if((phrases=fopen(argv[1], "r"))==NULL) {
		fprintf(stderr, "%s: Cannot open %s\n", argv[0], argv[1]);
		exit(2);
	}
	if((ideafile=fopen(argv[2], "r"))==NULL) {
		fprintf(stderr, "%s: Cannot open %s\n", argv[0], argv[2]);
		exit(3);
	}
	if(fread(&ctb, 1, 1, ideafile)==0) {
		fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
		exit(5);
	}
	ctb_return = ctb_check(ctb);
	switch(ctb_return) {
		case(1)  : secret_key = 1; break;
		case(-2) : fprintf(stderr, "%s: %s is not a PGP file\n", argv[0], argv[2]); exit(5);
		case(2)  : fprintf(stderr, "%s: %s is not conventionally encrypted\n", argv[0], argv[2]); exit(6);
		case(-1) : fprintf(stderr, "%s: %s appears to be ascii-armored\n", argv[0], argv[2]); exit(5);
		default  : break;
	}
/* Open a file to which to write the cracked passphrase.  If none is specified,
 * use stderr.
 */
	if(argv[3]) {
		if((log=fopen(argv[3], "w"))==NULL) {
			fprintf(stderr, "%s: Warning: cannot open logfile %s.  Using stderr.\n", argv[0], argv[3]);
			log=stderr;
		}
	} else {
		log=stderr;
	}
	if(secret_key==0) {
		if(fseek(ideafile, 5, SEEK_SET)==-1) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(fread(buf1, 1, 10, ideafile) < 10) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(fread(buf2, 1, 8, ideafile) < 8) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		fclose(ideafile);
		while((fgets(data, 256, phrases))!=NULL) {
			chop(data);
			hash(data, md5hash);
			if(ideacrack(buf1, buf2, md5hash)==0)
				fprintf(log, "PGPCrack passphrase: %s\n", data);
		}
	} else {
		alg=check_alg(ideafile);
		if(alg != 1) {
			fprintf(stderr, "%s: %s is not encrypted using IDEA\n", argv[0], argv[2]);
			exit(6);
		}
		if(fread(iv, 1, 8, ideafile) < 0) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(read_mpi(d, ideafile) < 0) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(read_mpi(p, ideafile) < 0) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(read_mpi(q, ideafile) < 0) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(read_mpi(u, ideafile) < 0) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		if(fread(&checksum2, 2, 1, ideafile) < 1) {
			fprintf(stderr, "%s: %s is corrupt\n", argv[0], argv[2]);
			exit(6);
		}
		p_len = bit2byte(conv_word16(p));
		q_len = bit2byte(conv_word16(q));
		d_len = bit2byte(conv_word16(d));
		u_len = bit2byte(conv_word16(u));
		checksum2 = byteorder(checksum2);
		fclose(ideafile);
		while((fgets(data, 256, phrases))!=NULL) {
			chop(data);
			hash(data, md5hash);
			if(secret_crack(iv, p, q, d, u, p_len, q_len, d_len, u_len, md5hash)==checksum2)
				fprintf(log, "PGPCrack passphrase: %s\n", data);
		}
	}
	fclose(phrases);
	fclose(log);
	exit(0);
}

/* This function is roughly equivalent to the chop() function in Perl.  This
 * chops off the last byte of the character array and if the second to last
 * byte contains a '\r' character (0x0d), then that byte is also deleted.
 * Passphrase lists can now be in DOS text format also.
 */

void inline chop(char *buf)
{
	int i;
	i = strlen(buf);
	buf[i-1] = 0;
	if(buf[i-2] == 13)
		buf[i-2] = 0;
}

void inline hash(char *buf, char *md5hash)
{
	struct MD5Context md5;
	MD5Init(&md5);
	MD5Update(&md5, buf, strlen(buf));
	MD5Final(md5hash, &md5);
}

/* Check to make sure that byte is a valid CTB. */
int ctb_check(char c)
{
	int i;
	i = (int)c;
	if(i & 0x80) {
		if((i & 0x24)==36)
			return 0;
		else if((i & 0x14)==20)
			return 1;
		else
			return 2;
	}
	if(i == 0x2d)
		return -1;
	return -2;
}

int inline ideacrack(char *buf1, char *buf2, char *key)
{
	struct IdeaCfbContext *ctx; 
	char plainbuf1[10];
	char plainbuf2[8];
	ideaCfbInit(ctx, key); 
	ideaCfbDecrypt(ctx, buf1, plainbuf1, 10);
	if((plainbuf1[6] == plainbuf1[8]) && (plainbuf1[7] == plainbuf1[9])) {
		ideaCfbSync(ctx);
		ideaCfbDecrypt(ctx, buf2, plainbuf2, 8);
		if(ctb_check(plainbuf2[0]) < 0)
			return 1;
		return 0;
	}
	return 1;
}

word16 inline bit2byte(word16 word)
{
	return (word+7)/8;
}

word16 inline byteorder(word16 word)
{
#ifndef WORDS_BIGENDIAN
	return ((word >> 8) | (word << 8));
#else
	return word;
#endif
}

/* This function finds the algorithm identifier in a secret key file and sees
 * whether it is encrypted with IDEA or not.
 */

unsigned char check_alg(FILE *fp)
{
	unsigned char c;
	if(fseek(fp, HEADERLEN, SEEK_SET) < 0)
		return 255;
	if(skip_mpi(fp) < 0)
		return 255;
	if(skip_mpi(fp) < 0)
		return 255;
	if(fread(&c, 1, 1, fp) < 1)
		return 255;
	return c;
}

int skip_mpi(FILE *fp)
{
	word16 len;
	byte *buf;
	if(fread(buf, 1, 2, fp) < 2)
		return -1;
	len = bit2byte(conv_word16(buf));
	if(fseek(fp, len, SEEK_CUR) < 0)
		return -1;
	return 0;
}

/* Reads in a Multiple-Precision Integer from a secret key file, converting
 * it to the appropriate byte-order.
 */

int read_mpi(byte *dat, FILE *fp)
{
	word16 len;
	if(fread(dat, 1, 2, fp) < 2)
		return -1;
	len = conv_word16(dat);
	len = bit2byte(len);
	if(fread(dat+2, 1, len, fp) < len)
		return -1;
	return 0;
}

word16 inline secret_crack(byte *iv, byte p[512], byte q[512], byte d[512],
	byte u[512], word16 p_len, word16 q_len, word16 d_len, word16 u_len, char *key)
{
	struct IdeaCfbContext ctx;	
	byte test_iv[8];
	register int i;
	word16 csum = 0;
	byte test_p[512], test_q[512], test_d[512], test_u[512];
	test_p[0] = p[0]; test_p[1] = p[1];
	test_q[0] = q[0]; test_q[1] = q[1];
	test_d[0] = d[0]; test_d[1] = d[1];
	test_u[0] = u[0]; test_u[1] = u[1];
	ideaCfbInit(&ctx, key);
	ideaCfbDecrypt(&ctx, iv, test_iv, 8);
	ideaCfbSync(&ctx);
	ideaCfbDecrypt(&ctx, d+2, test_d+2, d_len);
	ideaCfbDecrypt(&ctx, p+2, test_p+2, p_len);
	ideaCfbDecrypt(&ctx, q+2, test_q+2, q_len);
	ideaCfbDecrypt(&ctx, u+2, test_u+2, u_len);
	for(i = 0; i < d_len+2; i++)
		csum += test_d[i];
	for(i = 0; i < p_len+2; i++)
		csum += test_p[i];
	for(i = 0; i < q_len+2; i++)
		csum += test_q[i];
	for(i = 0; i < u_len+2; i++)
		csum += test_u[i];
	return csum;
}

word16 conv_word16(byte *buf)
{
	word16 i, j;
	i = buf[0];
	j = buf[1];
	return (j + (i << 8));
}
